/* 1012, Fri 2 Jun 95

   FLOWHASH.C:  AU Meter, using hash tables for counts

   Copyright (C) 1992-1995 by Nevil Brownlee,
   Computer Centre,  The University of Auckland */

#define DEBUG  /* Required if any of the following are set */

#define STACK_CHECK  /* Keep an eye on pattern & return stack extents */
#define noGC_TEST  /* Monitor garbage collector */

#define noMATCH_TRACE  /* Testing, testing .. */
#define noCOMPARE_TRACE
#define noP_TRACE
#define noGC_DEBUG
#define noTESTING
#define noGC_TEST1
#define noH_TEST

#define noDECNET

#include <stdio.h>

#ifndef AU_MSDOS
#include <malloc.h>

#else  /* AU_MSDOS */
#include <alloc.h>
#include <mem.h>
#ifdef DEBUG
#include <dos.h>
#endif
#endif

#include "ausnmp.h"

#include <sys/types.h>
#include <string.h>

#include "pktsnap.h"

#define EXTFLOW
#include "flowhash.h"
#include "decnet.h"

#ifdef DEBUG
void paddr(unsigned char far *a);
void daddr(unsigned char far *a);
void dport(unsigned char far *a);
void pkey(char *msg,struct flow far *n);
#endif

FILE *log;

#define FBLKSZ 425

struct flow far *fl INIT(NULL);
struct flow huge *fa;  int nf_flows INIT(0);

struct flow far *get_flow(void)  /* Get an unused flow */
{
   struct flow far *q;
   if (fl != NULL) {
      q = fl;  fl = fl->link.next;
      }
   else {
      if (nf_flows == 0) {
	 fa = (struct flow far *)farmalloc(sizeof(struct flow)*FBLKSZ);
	 if (fa == NULL) {
	    scpos(0,24);  printf("No memory for new flows!");
	    return NULL;
	    }
	 nf_flows = FBLKSZ-1;
	 }
      else {
	 ++fa;  --nf_flows;
	 }
#ifdef GC_DEBUG
scpos(0,24);  printf("\nfa=%lu, nff=%d, empty_ix=%u\n",fa,nf_flows,empty_ix);
#endif
      q = fa;
      }
   return q;
   }

void free_flow(struct flow far *q)  /* Return a flow to free list */
{
#ifdef GC_DEBUG
scpos(0,24);  printf("\nfree_flow(%lu), empty_ix=%u\n", q,empty_ix);
#endif
   q->link.next = fl;
   fl = q;
   }

#ifdef DEBUG
int log_nbr = 1;

void open_log()
{
   char fn[30];
   sprintf(fn,"flows.%03d",log_nbr++);
   log = fopen(fn, "w");
   }

void paddr(a)
unsigned char far *a;
{
   fprintf(log,"%u.%u.%u.%u", a[0],a[1],a[2],a[3]);
   }

void daddr(a)
unsigned char far *a;
{
   int j;
   printf("%u",a[0]);
/*   for (j = 1; j != RULE_ADDR_LEN; ++j) printf(".%u", a[j]); */
   for (j = 1; j != 4; ++j) printf(".%u", a[j]);
   }

void dport(a)
unsigned char far *a;
{
   printf("%u",a[0]  << 8 | a[1]);
   }

void pkey(msg,f)
char *msg;
struct flow far *f;
{
   if (!log) open_log();
   fprintf(log,"%s: %lu  %u\n   {", msg,f, f->fk.PeerAddrType);
   paddr(f->fk.Low.PeerAddress); fprintf(log," & ");
   paddr(f->fk.Low.PeerMask); fprintf(log,"} -> {");
   paddr(f->fk.High.PeerAddress); fprintf(log," & ");
   paddr(f->fk.High.PeerMask); fprintf(log,"}\n");
   }
#endif

#define DRT_SIZE  3  /* Default rule set: 3 rules, 1 count */
#define DCT_SIZE  1
struct rule default_rule_table[DRT_SIZE] = {
/* Selector,        Mask, Value,       Action, JumpIndex */
   {FTLOWPEERTYPE,
#ifdef BIG_PEER_ADDR
	   255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
      AT_DUMMY,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
#else
      255,0,0,0,0,0,   AT_DUMMY,0,0,0,0,0,
#endif
      RA_IGNORE, 0}, /* SourcePeerType & 255 = dummy: Ignore, 0; */
   {RS_NULL,
#ifdef BIG_PEER_ADDR
      0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
      0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
#else
      0,0,0,0,0,0,   0,0,0,0,0,0,
#endif
      RA_GOTOACT, 3},  /* Null & 0 = 0: GotoAct, Next */
   {FTLOWPEERTYPE,
#ifdef BIG_PEER_ADDR
      255,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
#else
      255,0,0,0,0,0,   0,0,0,0,0,0,
#endif
      RA_COUNTPKT, 1} };  /* SourcePeerType & 255 = 0: CountPkt, 1 */

unsigned int rule_table_counts(unsigned char t)
{
   unsigned int rsz, n, j, k;
   struct rule huge *rt, *rp;
   if (rt_size[t-1] == 0) return 0;  /* Incomplete tables */
   rt = rule_table[t-1];  rsz = rt_size[t-1];
   for (n = j = 0;  j != rsz;  ++j) {  /* Mark goto break points */
      rp = &rt[j];  k = rp->RuleAction;
      if (k == RA_COUNT || k == RA_COUNTPKT) {
         if (rp->RuleJumpIndex > n) n = rp->RuleJumpIndex;
         }
      }
   return n;
   }

void optimise_rule_table()
{
   unsigned int j, k, g, h, x, l,
      nht,  /* Nbr of rule_hash_tbls */
      ths,  /* Sum of rule_hash_tbl sizes */
      rhss;  /* Total rule_hash sequence size */
   struct rule huge *rp, far *hp;
   struct rule_hash_tbl huge *c_rh;
   unsigned short hash;

   hp = &c_rt[0];  hp->hash_tbl_index = 1;
#ifdef H_TEST
   scpos(0,24);  printf("\nRT: c_rsz=%u, RULE_ADDR_LEN=%d\n",
      c_rsz,RULE_ADDR_LEN);
#endif
   for (j = 1;  j != c_rsz;  ++j) {  /* Mark match break points */
      rp = &c_rt[j];
      rp->hash_tbl_index = rp->RuleSelector != hp->RuleSelector ||
	    qcmp(rp->RuleMask,hp->RuleMask,RULE_ADDR_LEN) != 0;
      hp = rp;
      }
   for (j = 0;  j != c_rsz;  ++j) {  /* Mark goto break points */
      rp = &c_rt[j];  rp->hash_link = 0;
      k = rp->RuleAction;
      if (k == RA_PUSHTO || k == RA_PUSHTOACT
	    || k == RA_PUSHPKTTO || k == RA_PUSHPKTTOACT
	    || k == RA_GOTO || k == RA_GOTOACT
	    || k == RA_GOSUB || k == RA_GOSUBACT
	    || k == RA_ASSIGN || k == RA_ASSIGNACT
#ifdef POP_ACT
	    || k == RA_POPTO || k == RA_POPTOACT
#endif
	    ) {
	 hp = &c_rt[rp->RuleJumpIndex-1];
	 hp->hash_tbl_index = 2;
	 }
      }
#ifdef H_TEST
/*   scpos(0,24);  printf("rule  index  link  MARKED\n"); */
   scpos(0,24);  printf("rule  index  link  MARKED\n");
   for (j = 0;  j != c_rsz;  ++j) {
      rp = &c_rt[j];
      printf("%3d %5d %5d  %d  ",
	 j,rp->hash_tbl_index,rp->hash_link, rp->RuleSelector);
      daddr(rp->RuleMask);  printf("\n");
      }
#endif
   nht = ths = 0;
   hp = &c_rt[0];  k = 1;
   for (j = 1;  j != c_rsz;  ++j) {  /* Find compare groups */
      rp = &c_rt[j];
      if ((g = rp->hash_tbl_index) == 0) ++k;  /* Same group */
      if (g != 0 || j == c_rsz-1) {  /* End of a group */
	 if (k >= MNCGRPSZ && hp->RuleSelector != RS_NULL) {
	    for (h = 8; h < k; h <<= 1) ;  /* h = next power of 2 */
	    if (h > MXRHTSZ) h = MXRHTSZ;
	    hp->hash_tbl_index = k;  /* Remember group and */
	    hp->hash_link = h;       /*   hash table sizes */
	    ++nht;  ths += h;
	    }
	 else hp->hash_tbl_index = 0;
	 hp = rp;  k = 1;
	 }
      rp->hash_tbl_index = 0;  /* Last row can't start a group */
      }
   rhss = 1 + nht*2 + ths;  /* rule_hash set size (unsigned shorts) */
   if (rh_size < rhss) {
      farfree(rule_hash);
      rule_hash = (int far *)farmalloc(sizeof(int)*(rh_size = rhss));
      }
#ifdef H_TEST
   printf("rule  index  link  GROUPED\n");
   for (j = 0;  j != c_rsz;  ++j) {
      rp = &c_rt[j];
      printf("%3d %5d %5d\n", j, rp->hash_tbl_index,rp->hash_link);
      }
#endif
   rule_hash[0] = 0;  k = 1;  rp = &c_rt[j = 0];
   for ( ; nht != 0; --nht, k += 2+h) {
      c_rh = (struct rule_hash_tbl far *)&rule_hash[k];
      while (rp->hash_tbl_index == 0) {
	 ++rp; ++j;
	 }
      g = rp->hash_tbl_index;  /* Group size */
      h = rp->hash_link;  /* Hash table size */
      rp->hash_tbl_index = k;  /* Index of hash_tbl in rule_hash */
      rp->hash_link = 0;
      c_rh->group_last = j+g;
      c_rh->hash_mask = h-1;
      for (x = 0; x != h; ++x) c_rh->hash_ent[x] = 0;
      for ( ; g != 0; --g) {
	 for (hash = x = 0; x != RULE_ADDR_LEN; ++x)
	    hash += rp->RuleMatchedValue[x] & rp->RuleMask[x];
	 hash &= h-1;
	 if ((l = c_rh->hash_ent[hash]) == 0)
	    c_rh->hash_ent[hash] = j+1;  /* First rule with this hash */
	 else {
	    do {
	       hp = &c_rt[l-1];  l = hp->hash_link;
	       } while (l != 0);
	    hp->hash_link = j+1;  /* Add to end of hash chain */
	    }
	 ++rp; ++j;
	 }
      }
#ifdef H_TEST
   printf("Rule Hash Tables ..\n\nrule index  n_rules n_hash n_zero\n\n");
   for (j = 0;  j != c_rsz;  ++j) {
      rp = &c_rt[j];
      if ((k = rp->hash_tbl_index) == 0) continue;
      c_rh = (struct rule_hash_tbl far *)&rule_hash[k];
      h = c_rh->hash_mask+1;
      for (g = 0, hash = 0; hash != h; ++hash)
	 if (c_rh->hash_ent[hash] == 0) ++g;
      printf("%5d%5d  %5d%5d%5d\n     ",
	 j+1, k, c_rh->group_last-j+1, h, g);
      printf("\n");
      }
#endif
   }

int open_rule_set(unsigned char doSet, unsigned char n)
{
   unsigned int n_counts, j,k,p;
   struct rule far *rp;
   struct flow far *ap;
   if (rt_size[n-1] == 0) return 0;  /* Incomplete tables */
   if ((n_counts = rule_table_counts(n)) > MXCTBLS)
      return 0;  /* Too many counts */
   if (doSet) {
      ct_size[n-1] = c_csz = n_counts;
      for (k = 0; k != sizeof(proto_reqd); ++k) proto_reqd[k] = 0;
      adj_reqd = 0;
      c_rt = rule_table[n-1];  c_rsz = rt_size[n-1];
      for (j = 0; j != c_rsz; ++j) {
	 rp = &c_rt[j];
	 if (rp->RuleSelector == FTLOWPEERTYPE ||
	       rp->RuleSelector == FTHIPEERTYPE) {
	    p = rp->RuleMatchedValue[0];
	    if (rp->RuleAction == RA_PUSHPKTTO ||  /* All peer types */
		  rp->RuleAction == RA_PUSHPKTTOACT ||
		  rp->RuleAction == RA_COUNTPKT)
	       for (k = 0; k != sizeof(proto_reqd); ++k) proto_reqd[k] = 1;
	    else if (p <= AT_OTHER & *proto_names[p] != '\0')
	       proto_reqd[p] = 1;  /* Specific peer type */
	    }
	 if (rp->RuleSelector == FTLOWADJACENTADDRESS ||
	       rp->RuleSelector == FTHIADJACENTADDRESS)
	    adj_reqd = 1;
	 }
      if (adj_reqd) {
	 for (k = 0; k != sizeof(proto_reqd); ++k)
	    if (proto_reqd[k]) break;
	 if (k == sizeof(proto_reqd))  /* No protocol specified! */
	    for (k = 0; k != sizeof(proto_reqd); ++k) proto_reqd[k] = 1;
	 }
      CurrentRuleSet = n;
      optimise_rule_table();
      }
   return 1;
   }

void close_rule_set(void)
{
   unsigned int j,k;
   struct rule far *rp;
   struct flow far *fp, far *ap, far *lfp;
   struct hash_tbl huge *ht;
   struct flow far * far *sf;
   for (j = 0; j != c_rsz; ++j) {
      rp = &c_rt[j];
      if (rp->RuleAction == RA_COUNT ||
	    rp->RuleAction == RA_COUNTPKT) {
	 ht = count_table[rp->RuleJumpIndex-1];
	 if (ht != NULL) {
	    for (k = 0; k != HASHMOD; ++k) {
	       sf = &ht->hash_ent[k];  /* Address of hash table entry */
	       if ((fp = sf[0]) != NULL) {
		  do {
		     lfp = fp;  fp = lfp->link.next;
		     lfp->link.next = NULL;
		     } while (fp != (struct flow far *)sf);
		  --n_hash_ents;
		  }
	       }
	    farfree(ht);
	    --n_hash_tables;  t_hash_size -= HASHMOD;
	    count_table[rp->RuleJumpIndex-1] = NULL;
	    }
	 }
      }
   }

#ifdef GC_DEBUG
int dup_flow(struct flow far *f, char *msg)
{
   int x;
   for (x = 253; x <= mxflowsp1; ++x)
      if (flow_ix[x-1] != NULL) {
	 scpos(0,24);
	 printf("   flow_ix[%d]=%lu !!\n", x,flow_ix[x-1]);
	 }
   for (x = 2; x <= mxflowsp1; ++x)
      if (flow_ix[x-1] == f) {
	 scpos(0,24);
	 printf("Flow %lu is flow_ix[%d], msg=%s\n", f,x, msg);
	 exit(0);
	 }
   return 0;
   }
#endif

unsigned int number_flow(struct flow far *fp)  /* Allocate a flow_nbr */
{
   unsigned int start_ix = empty_ix;
   do {
      if (++empty_ix > mxflowsp1)  /* empty_ix goes from 2 to MXFLOWS+1 */
	 empty_ix = 2;  /* Flow 1 is a dummy for snmp */
      if (flow_ix[empty_ix-1] == NULL) {
	 ++nflows;  /* Nbr of active accounting flows */
	 ++FlowsCreated;
	 flow_ix[empty_ix-1] = fp;
#ifdef GC_DEBUG
	 printf("nbr_flow(%lu) -> %u\n", fp,empty_ix);
#endif
	 return empty_ix;  /* 1-org */
	 }
      } while (empty_ix != start_ix);

   if (garbage_collect(0)) {  /* No unused flows: force garbage collection */
      ++nflows;  /* Nbr of active accounting flows */
      ++FlowsCreated;
      flow_ix[gcf_ix-1] = fp;
      return empty_ix = gcf_ix;  /* 1-org */
      }
   else {
      more_garbage();  /* Increase garbage collecting rate */
      scpos(0,24);  printf("Too many flows");
      return 0;
      }
   }

struct flow far *find_flow(int x)  /* Find flow with flow_nbr x */
{
   if (x > mxflowsp1 || x < 2) {
      display_msg(1,"Invalid flow number!");
      return NULL;
      }
   return flow_ix[x-1];
   }

void more_garbage()
{
   unsigned int max_gc_f;
   gc_interval >>= 1;  /* Decrease garbage collection interval */
   if (gc_interval < 2) {
      gc_interval = 2;
      max_gc_f = (mxflowsp1-1)/5;
      gc_f <<= 1;  /* Increase nbr of flows checked each collection */
      if (gc_f > max_gc_f) gc_f = max_gc_f;
      }
   }

int garbage_collect(unsigned char incremental)
{
   unsigned char f,e;
   unsigned int s_gcf_ix;
   struct rule far *rp;
   struct flow far *fp, far *np, far *tp, far *ltp;
#ifdef GC_DEBUG
   int k;
#endif
#ifdef GC_TEST1
   scpos(0,24);
   printf("GC: gctime=%lu, gcf_ix=%u, e=%d, f=%d\n",
      GarbageCollectTime,gcf_ix,e,f);
#endif
   e = gc_f;  f = gc_f/4;
   s_gcf_ix = gcf_ix;
   do {
      if (++gcf_ix > mxflowsp1)  /* gcf_ix goes from 2 to MXFLOWS+1 */
	 gcf_ix = 2;  /* Flow 1 is a dummy for snmp */
      if ((fp = flow_ix[gcf_ix-1]) == NULL) {  /* Unused flow index */
	 if (incremental && --e == 0) return 1;
	 continue;
         }
      if (fp->LastTime <= GarbageCollectTime) {  /* OK, recover it */
#ifdef GC_TEST1
      printf("   f=%d, gcf_ix=%d, LastTime=%lu\n",
	 f,gcf_ix,fp->LastTime);
#endif
	 if (fp->link.next != NULL) {  /* Not an orphaned flow */
#ifdef GC_DEBUG
	    scpos(0,24);
	    printf("GC: set=%u, ix=%u, fp=%lu, np=%lu\n",
	       fp->FlowRuleSet,gcf_ix, fp,fp->link.next);
	    k = 0;
#endif
	    np = fp->link.next;
	    tp = fp;  do {
	       ltp = tp;
#ifdef GC_DEBUG
printf("   k=%d, ltp=%lu, tp=%lu\n", ++k, ltp,ltp->link.next);
if (tp->link.next == NULL || k > 10) exit(0);  /* Infinite loop - stop */
#endif
	       } while ((tp = tp->link.next) != fp);
	    ltp->link.next = ltp == np ? NULL : np;
	    }
	 flow_ix[gcf_ix-1] = NULL;
#ifdef GC_DEBUG
      printf("\ngarbage_collect ..\n");
#endif
	 free_flow(fp);
	 --nflows;
	 ++FlowsRecovered;
	 if (!incremental) return 1;  /* Forced collection succeeded,
					    flow gcf_ix is now free */
#ifdef GC_TEST
      scpos(0,24);
      printf("Flow %d recovered\n", gcf_ix);
#endif
	 }
      if (incremental && --f == 0) return 1;
      } while (gcf_ix != s_gcf_ix);
   return 0;  /* Failed to recover any flows */
   }

struct flow far *tally;  /* Work space for building target flows */

unsigned int p_stack[RSTKSIZ];  /* Pattern stack */
unsigned char a_stack[RSTKSIZ];  /* Attribute values for pattern stack */
int p_s_depth;

unsigned char user_vbls[NBRVBLS];  /* User variables */
unsigned int env_stack[ESTKSIZ];  /* Environment stack */
int e_s_tos;

#ifdef GC_DEBUG
void print_chain(struct hash_tbl huge *t, unsigned short hash, char *msg)
{
   struct flow far * far *sf;
   struct flow far *f, far *lf;
   sf = &t->hash_ent[hash];  /* Address of hash table entry */
   scpos(0,24);  printf("Chain: t=%lu, hash=%u, sf=%lu, msg=%s\n",
      t,hash,sf, msg);
   lf = f = sf[0];
   do {
      printf("   f=%lu\n", f);
      if ((lf = f) == NULL) return;
      } while ((f = lf->link.next) != (struct flow far *)sf);
   }
#endif

struct flow far *search_hash_table(
   struct hash_tbl far *t,  /* Table to search */
   struct pkt_key far *Low,  /* Target keys */
   struct pkt_key far *High,
   struct pkt far *bp,
   unsigned char OK_to_add  /* 1 = add new flow if not found */
   )
{
   struct flow far * far *sf;
   struct flow far *f, far *lf;
   struct rule far *rp;
   unsigned short hash;
   unsigned char j, k, pal,
      t_LowAdj,t_HighAdj, t_LowPeer,t_HighPeer, t_LowDetail,t_HighDetail,
      push_rule;
   t_LowAdj = t_HighAdj = t_LowPeer = t_HighPeer =
      t_LowDetail = t_HighDetail = hash = 0;
   ++n_hash_searches;
   memcpy((unsigned char far *)&tally->fk,  /* Start with a null flow */
      (unsigned char far *)&null_flow->fk, sizeof(struct flow_key));
   pal = PEER_ADDR_LEN;  /* Could be any type, start by using max length */
#ifdef TESTING
   if (!log) open_log();
   fprintf(log,"p_s_depth=%d\n", p_s_depth);
#endif
   for (j = 0; j != p_s_depth; ++j) {
      rp = &c_rt[p_stack[j]-1];
#ifdef TESTING
      fprintf(log,"   j=%d, attrib=%d\n", j,a_stack[j]);
#endif
      push_rule = (rp->RuleAction != RA_PUSHPKTTO) &&
	  (rp->RuleAction != RA_PUSHPKTTOACT) &&
	  (rp->RuleAction != RA_COUNTPKT);
      switch (a_stack[j]) {
      case FTLOWPEERADDRESS:
	 t_LowPeer = 1;  /* Tally this field */
	 if (push_rule) for (k = 0; k != pal; ++k)
	    hash += (tally->fk.Low.PeerAddress[k] =
	       rp->RuleMatchedValue[k]);
	 else for (k = 0; k != pal; ++k)
	    hash += (tally->fk.Low.PeerAddress[k] =
	       Low->PeerAddress[k] & rp->RuleMask[k]);
	 break;
      case FTHIPEERADDRESS:
	 t_HighPeer = 1;  /* Tally this field */
	 if (push_rule) for (k = 0; k != pal; ++k)
	    hash += (tally->fk.High.PeerAddress[k] =
	       rp->RuleMatchedValue[k]);
	 else for (k = 0; k != pal; ++k)
	    hash += (tally->fk.High.PeerAddress[k] =
	       High->PeerAddress[k] & rp->RuleMask[k]);
	 break;
      case FTLOWPEERTYPE:
      case FTHIPEERTYPE:  /* PeerType allways tallied */
	 if (push_rule)
	    hash += (tally->fk.PeerAddrType = rp->RuleMatchedValue[0]);
	 else hash +=
	    (tally->fk.PeerAddrType = bp->PeerAddrType);
	 pal = addr_len[tally->fk.PeerAddrType];  /* Now we know peer type */
	 break;
      case FTLOWDETAILADDRESS:
	 t_LowDetail = 1;  /* Tally this field */
	 if (push_rule) for (k = 0; k != DETAIL_ADDR_LEN; ++k)
	    hash += (tally->fk.Low.DetailAddress[k] =
	       rp->RuleMatchedValue[k]);
	 else for (k = 0; k != DETAIL_ADDR_LEN; ++k)
	    hash += (tally->fk.Low.DetailAddress[k] =
	       Low->DetailAddress[k] & rp->RuleMask[k]);
	 break;
      case FTHIDETAILADDRESS:
	 t_HighDetail = 1;  /* Tally this field */
	 if (push_rule) for (k = 0; k != DETAIL_ADDR_LEN; ++k)
	    hash += (tally->fk.High.DetailAddress[k] =
	       rp->RuleMatchedValue[k]);
	 else for (k = 0; k != DETAIL_ADDR_LEN; ++k)
	    hash += (tally->fk.High.DetailAddress[k] =
	       High->DetailAddress[k] & rp->RuleMask[k]);
	 break;
      case FTLOWDETAILTYPE:
      case FTHIDETAILTYPE:  /* DetailType always tallied */
	 if (push_rule)
	    hash += (tally->fk.DetailAddrType = rp->RuleMatchedValue[0]);
	 else hash +=
	    (tally->fk.DetailAddrType = bp->DetailAddrType);
	 break;
      case FTLOWADJACENTADDRESS:
	 t_LowAdj = 1;  /* Tally this field */
	 if (push_rule) for (k = 0; k != MAC_ADDR_LEN; ++k)
	    hash += (tally->fk.Low.AdjAddress[k] =
	       rp->RuleMatchedValue[k]);
	 else for (k = 0; k != MAC_ADDR_LEN; ++k)
	    hash += (tally->fk.Low.AdjAddress[k] =
	       Low->AdjAddress[k] & rp->RuleMask[k]);
	 break;
      case FTHIADJACENTADDRESS:
	 t_HighAdj = 1;  /* Tally this field */
	 if (push_rule) for (k = 0; k != MAC_ADDR_LEN; ++k)
	    hash += (tally->fk.High.AdjAddress[k] =
	       rp->RuleMatchedValue[k]);
	 else for (k = 0; k != MAC_ADDR_LEN; ++k)
	    hash += (tally->fk.High.AdjAddress[k] =
	       High->AdjAddress[k] & rp->RuleMask[k]);
	 break;
      case FTSOURCECLASS:
	 hash += (tally->fk.Low.Class = rp->RuleMatchedValue[0]);
	 break;
      case FTDESTCLASS:
	 hash += (tally->fk.High.Class = rp->RuleMatchedValue[0]);
	 break;
      case FTFLOWCLASS:
	 hash += (tally->fk.FlowClass = rp->RuleMatchedValue[0]);
	 break;
      case FTSOURCEKIND:
	 hash += (tally->fk.Low.Kind = rp->RuleMatchedValue[0]);
	 break;
      case FTDESTKIND:
	 hash += (tally->fk.High.Kind = rp->RuleMatchedValue[0]);
	 break;
      case FTFLOWKIND:
	 hash += (tally->fk.FlowKind = rp->RuleMatchedValue[0]);
	 break;
	 }
      }
#ifdef MATCH_TRACE
   printf("\n  Type,Class,Kind: %d  %d %d %d  %d %d %d  hash=%d|",
      tally->fk.PeerAddrType,
      tally->fk.Low.Class,tally->fk.High.Class,tally->fk.FlowClass,
      tally->fk.Low.Kind,tally->fk.High.Kind,tally->fk.FlowKind,
      hash);
#endif
   sf = &t->hash_ent[hash &= HASHMOD-1];  /* Address of hash table entry */
#ifdef TESTING
if (!log) open_log();
fprintf(log,
"Search_hash(): add=%d, lo=%d, hi=%d, hash=%d\n   Type=%d, LowPeer=",
   OK_to_add, t_LowPeer,t_HighPeer, hash,
   tally->fk.PeerAddrType);
paddr(tally->fk.Low.PeerAddress); fprintf(log,", HighPeer=");
paddr(tally->fk.High.PeerAddress);
fprintf(log,"\n   sf=%lu, sf[0]=%lu\n", sf,sf[0]);
#endif
   if ((f = sf[0]) == NULL) {
      lf = NULL;  /* Empty hash chain */
      ++n_hash_compares;  /* Make sure compares >= searches! */
      }
   else {
      do {
#ifdef GC_DEBUG
	 if (f == NULL) {
	    print_chain(t,hash,"existing");
	    exit(0);
	    }
#endif
	 ++n_hash_compares;
	 for (;;) {  /* Compare whole fields - data has been masked */
	    if (t_LowPeer && qcmp(tally->fk.Low.PeerAddress,
	       f->fk.Low.PeerAddress,pal) != 0) break;
	    if (t_HighPeer && qcmp(tally->fk.High.PeerAddress,
	       f->fk.High.PeerAddress,pal) != 0) break;
	    if (tally->fk.PeerAddrType != f->fk.PeerAddrType) break;
#ifdef COMPARE_TRACE
	    printf("\n  hash_cmp: pal=%u  %u %u   ",
	       pal, t_LowDetail,t_HighDetail);
	    printf("%u:%u ",tally->fk.PeerAddrType,f->fk.PeerAddrType);
	    dport(tally->fk.Low.DetailAddress);  printf(":");
	    dport(f->fk.Low.DetailAddress);  printf("  ");
	    dport(tally->fk.High.DetailAddress);  printf(":");
	    dport(f->fk.High.DetailAddress);  printf("|");
#endif
	    if (t_LowDetail && qcmp(tally->fk.Low.DetailAddress,
	       f->fk.Low.DetailAddress,DETAIL_ADDR_LEN) != 0) break;
	    if (t_HighDetail && qcmp(tally->fk.High.DetailAddress,
	       f->fk.High.DetailAddress,DETAIL_ADDR_LEN) != 0) break;
	    if (tally->fk.DetailAddrType != f->fk.DetailAddrType) break;

	    if (t_LowAdj && qcmp(tally->fk.Low.AdjAddress,
	       f->fk.Low.AdjAddress,MAC_ADDR_LEN) != 0) break;
	    if (t_HighAdj && qcmp(tally->fk.High.AdjAddress,
	       f->fk.High.AdjAddress,MAC_ADDR_LEN) != 0) break;

	    if (tally->fk.Low.Class != f->fk.Low.Class) break;
	    if (tally->fk.High.Class != f->fk.High.Class) break;
	    if (tally->fk.FlowClass != f->fk.FlowClass) break;
	    if (tally->fk.Low.Kind != f->fk.Low.Kind) break;
	    if (tally->fk.High.Kind != f->fk.High.Kind) break;
	    if (tally->fk.FlowKind != f->fk.FlowKind) break;
#ifdef MATCH_TRACE
   printf("\n  matched, hash=%u, f=%lu|", hash,f);
#endif
	    return f;  /* Matched */
	    }
	 lf = f;
	 } while ((f = lf->link.next) != (struct flow far *)sf);
      f = NULL;  /* Loop stops with lf -> last flow in chain */
      }
#ifdef TESTING
fprintf(log,"   after search: f=%lu, lf=%lu, sf=%lu, sf[0]=%lu\n",
   f, lf, sf, sf[0]);
#endif
   if (OK_to_add) {
      if ((f = get_flow()) == NULL) {  /* Build new count flow */
	 ++noflowpackets;
	 return NULL;  /* Out of memory */
	 }
      memcpy((unsigned char far *)&f->fk,
	 (unsigned char far *)&tally->fk, sizeof(struct flow_key));
#ifdef MATCH_TRACE
   printf("\n  new flow, hash=%u  %d %d  %d %d:  f=%lu\n", hash,
      t_LowPeer,t_HighPeer, t_LowDetail,t_HighDetail, f);
   printf("Count: %u+%u",a_stack[0],p_stack[0]);
   for (j = 1; j != p_s_depth; ++j) printf(",%u+%u",a_stack[j],p_stack[j]);
   printf(": ");
   daddr(Low->PeerAddress);  printf("->");
   daddr(High->PeerAddress);  printf("\n");
#endif
      f->FirstTime = s_uptime;  /* Initialise flow */
      f->UpOctets = f->UpPDUs = f->DownOctets = f->DownPDUs =
	 f->LastTime = 0L;
      f->FlowRuleSet = CurrentRuleSet;
      f->link.next = NULL;
      for (j = 0; j != p_s_depth; ++j) {  /* Set masks and rule-derived */
	 rp = &c_rt[p_stack[j]-1];        /*    attribute values */
	 switch (a_stack[j]) {
	 case FTLOWPEERADDRESS:
	    addrcpy(f->fk.Low.PeerMask, rp->RuleMask, pal);
	    break;
	 case FTHIPEERADDRESS:
	    addrcpy(f->fk.High.PeerMask, rp->RuleMask, pal);
	    break;
	 case FTLOWDETAILADDRESS:
	    addrcpy(f->fk.Low.DetailMask, rp->RuleMask, DETAIL_ADDR_LEN);
	    break;
	 case FTHIDETAILADDRESS:
	    addrcpy(f->fk.High.DetailMask, rp->RuleMask, DETAIL_ADDR_LEN);
	    break;
	 case FTLOWADJACENTADDRESS:
	    addrcpy(f->fk.Low.AdjMask, rp->RuleMask, MAC_ADDR_LEN);
	    break;
	 case FTHIADJACENTADDRESS:
	    addrcpy(f->fk.High.AdjMask, rp->RuleMask, MAC_ADDR_LEN);
	    break;
	    }
	 }
      if (number_flow(f) == 0) {  /* No room in flow_index */
	 free_flow(f);
	 ++noflowpackets;
	 return NULL;
	 }
      f->link.next = (struct flow far *)sf;
#ifdef GC_DEBUG
      if (f->link.next == NULL) {
	 print_chain(t,hash,"adding new flow");
	 exit(0);
	 }
#endif
      if (lf == NULL) {  /* Empty hash chain */
	 sf[0] = f;  ++n_hash_ents;
	 }
      else lf->link.next = f;
#ifdef TESTING
fprintf(log,"   flow %d added, addr=%lu\n", empty_ix,f);
#endif
      }
   return f;
   }

unsigned char masked_value[RULE_ADDR_LEN];

unsigned short mk_hash(
   unsigned char far *vp, unsigned char far *mp, unsigned char n)
{
   unsigned short hash;
   unsigned char j;
   for (j = hash = 0;  j != n;  ++j)
      hash += (masked_value[j] = vp[j] & mp[j]);
   return hash;
   }

int pkt_match(  /* 0 = failed to match,  1 = matched, */
		/* 2 = matched but flow not yet in count table */
   unsigned char forward,  /* Value of meter's 'Forward' attribute */
   unsigned char OK_to_add,  /* Create the flow if possible */
   unsigned char flow_order,   /* Increment 'up' (source->dest) counts */
   struct pkt far *bp, struct pkt_key far *from, struct pkt_key far *to)
{
   unsigned int rx, j, f;
   struct rule far *rp;
   unsigned short hash;
   unsigned char attrib, match, test_reqd, pal,
      ral;  /* Rule address length */
   struct flow far *fp, far *ap;
   struct rule_hash_tbl far *c_rh;
   struct hash_tbl far *ht;

   e_s_tos = p_s_depth = 0;  test_reqd = 1;
   pal = PEER_ADDR_LEN;
   for (rx = 1; rx <= c_rsz; ) {  /* Search the rule table */
      rp = &c_rt[rx-1];
      if ((attrib = rp->RuleSelector) >= FTV1)  /* User variable */
	 /* Dereference it: this is why we need a_stack! */
	 attrib = user_vbls[rp->RuleSelector-FTV1];
      if (test_reqd) {
	 if (bp->PeerAddrType != AT_DUMMY) ++n_matches;
	 if (rp->hash_tbl_index == 0) {  /* Simple compare */
	    switch (attrib) {
	    case FTLOWPEERADDRESS:
	       for (j = 0;  j != pal;  ++j) {
		  if ((from->PeerAddress[j] & rp->RuleMask[j]) !=
		     rp->RuleMatchedValue[j]) break;
		  }
	       match = j == pal;
	       break;
	    case FTHIPEERADDRESS:
	       for (j = 0;  j != pal;  ++j) {
		  if ((to->PeerAddress[j] & rp->RuleMask[j]) !=
		     rp->RuleMatchedValue[j]) break;
		  }
	       match = j == pal;
	       break;
	    case FTLOWPEERTYPE:
	    case FTHIPEERTYPE:
	       match = bp->PeerAddrType == rp->RuleMatchedValue[0];
	       if (match) pal = addr_len[bp->PeerAddrType];
	       break;
	    case FTLOWDETAILADDRESS:
	       for (j = 0;  j != DETAIL_ADDR_LEN;  ++j) {
		  if ((from->DetailAddress[j] & rp->RuleMask[j]) !=
		     rp->RuleMatchedValue[j]) break;
		  }
	       match = j == DETAIL_ADDR_LEN;
	       break;
	    case FTHIDETAILADDRESS:
	       for (j = 0;  j != DETAIL_ADDR_LEN;  ++j) {
		  if ((to->DetailAddress[j] & rp->RuleMask[j]) !=
		     rp->RuleMatchedValue[j]) break;
		  }
	       match = j == DETAIL_ADDR_LEN;
	       break;
	    case FTLOWDETAILTYPE:
	    case FTHIDETAILTYPE:
	       match = bp->DetailAddrType == rp->RuleMatchedValue[0];
	       break;
	    case FTLOWADJACENTADDRESS:
	       for (j = 0;  j != MAC_ADDR_LEN;  ++j) {
		  if ((from->AdjAddress[j] & rp->RuleMask[j]) !=
		     rp->RuleMatchedValue[j]) break;
		  }
	       match = j == MAC_ADDR_LEN;
	       break;
	    case FTHIADJACENTADDRESS:
	       for (j = 0;  j != MAC_ADDR_LEN;  ++j) {
		  if ((to->AdjAddress[j] & rp->RuleMask[j]) !=
		     rp->RuleMatchedValue[j]) break;
		  }
	       match = j == MAC_ADDR_LEN;
	       break;
	    case FTFORWARD:  /* Addresses in packet order */
	       match = forward == rp->RuleMatchedValue[0];
	       break;
	    case RS_NULL:
	    default:
	       match = 1;  /* Don't test anything */
	       }
	    }
	 else {  /* Hashed compare */
	    match = 0;
	    switch (attrib) {
	    case FTLOWPEERADDRESS:
	       hash = mk_hash(from->PeerAddress,rp->RuleMask,ral=pal);
	       break;
	    case FTHIPEERADDRESS:
	       hash = mk_hash(to->PeerAddress,rp->RuleMask,ral=pal);
	       break;
	    case FTLOWPEERTYPE:
	    case FTHIPEERTYPE:
	       hash = mk_hash(&bp->PeerAddrType,rp->RuleMask,ral=1);
	       break;
	    case FTLOWDETAILADDRESS:
	       hash = mk_hash(from->DetailAddress,rp->RuleMask,
		  ral=DETAIL_ADDR_LEN);
	       break;
	    case FTHIDETAILADDRESS:
	       hash = mk_hash(to->DetailAddress,rp->RuleMask,
		  ral=DETAIL_ADDR_LEN);
	       break;
	    case FTLOWDETAILTYPE:
	    case FTHIDETAILTYPE:
	       hash = mk_hash(&bp->DetailAddrType,rp->RuleMask,ral=1);
	       break;
	    case FTLOWADJACENTADDRESS:
	       hash = mk_hash(from->AdjAddress,rp->RuleMask,ral=MAC_ADDR_LEN);
	       break;
	    case FTHIADJACENTADDRESS:
	       hash = mk_hash(to->AdjAddress,rp->RuleMask,ral=MAC_ADDR_LEN);
	       break;
	    case RS_NULL:
	    default:
	       match = 1;  /* Don't test anything */
	       }
	    if (!match) {
	       c_rh = (struct rule_hash_tbl far *)&rule_hash[rp->hash_tbl_index];
	       hash &= c_rh->hash_mask;
#ifdef P_TRACE
	       j = rx;
#endif
	       if ((rx = c_rh->hash_ent[hash]) != 0) {
		  do {
		     rp = &c_rt[rx-1];
		     if (qcmp((unsigned char far *)masked_value,
			   rp->RuleMatchedValue, ral) == 0) {
			match = 1;  break;  /* Matched */
			}
		     } while ((rx = rp->hash_link) != 0);
		  }
	       }
	    if (match == 0) rx = c_rh->group_last;  /* Not in group */
#ifdef P_TRACE
      printf("Hash cmp: rx=%d, att=%d, mtch=%d, m_rx=%d, act=%d, ix=%d\n",
	 j,attrib, match,rx, rp->RuleAction,rp->RuleJumpIndex);
#endif
	    }
	 }
      else {  /* No test required */
#ifdef P_TRACE
      printf("No test: rx=%d, attrib=%d, action=%d, index=%d\n",
	 rx,attrib, rp->RuleAction,rp->RuleJumpIndex);
#endif
	 test_reqd = 1;  /* Didn't test for this rule */
	 match = 1;
	 }
      if (match) {
	 switch (rp->RuleAction) {
	 case RA_IGNORE:
	    return 1;
	 case RA_RETRY:
	    return 0;
	 case RA_COUNT:
	 case RA_COUNTPKT:
	    if (attrib != 0 || rp->RuleMask[0] != 0) {
				      /* Count rule specified an attribute */
	       a_stack[p_s_depth] = attrib;  /* Remember last matched rule */
	       p_stack[p_s_depth++] = rx;
	       }
#ifdef COUNT_TRACE
	    if ((from->PeerAddress[2] == 4 && from->PeerAddress[3] == 33) ||
		 (to->PeerAddress[2] == 4 && to->PeerAddress[3] == 33)) {
	       printf("\nCount: %u+%u",a_stack[0],p_stack[0]);
	       for (j = 1; j != p_s_depth; ++j) printf(",%u+%u",
		  a_stack[j],p_stack[j]);
	       printf(" %u ", forward);
	       daddr(from->PeerAddress);  printf(":");
	       dport(from->DetailAddress);
	       printf(" %s ", flow_order ? "->" : "<-");
	       daddr(to->PeerAddress);  printf(":");
	       dport(to->DetailAddress);  printf("|");
	       }
#endif
	    ht = count_table[rp->RuleJumpIndex-1];
	    if (ht == NULL) {  /* No hash table yet */
	       ht = (struct hash_tbl far *)farmalloc(sizeof(struct hash_tbl));
	       for (j = 0; j != HASHMOD; ++j) ht->hash_ent[j] = NULL;
	       count_table[rp->RuleJumpIndex-1] = ht;
	       ++n_hash_tables;  t_hash_size += HASHMOD;
	       }
	    if ((fp = search_hash_table(ht, from,to,bp,OK_to_add)) == NULL)
	       return 2;  /* Not in tally, not OK to add it yet */
	    break;  /* Count this PDU */
	 case RA_RETURN:
	    --e_s_tos;
#ifdef STACK_CHECK
	    if (e_s_tos < 0) {
	       scpos(0,24);
	       printf("gosub oflo: depth=%d, stk=", e_s_tos);
	       for (j = 0; j != ESTKSIZ; ++j) printf("%u,", env_stack[j]);
	       printf("\n");
	       return 0;
	       }
#endif
#ifdef P_TRACE
      scpos(0,24);
      printf("Return: rx=%d, attrib=%d, gosub@%d, index=%d\n",
	 rx,attrib, env_stack[e_s_tos],rp->RuleJumpIndex);
#endif
	    rx = env_stack[e_s_tos] + rp->RuleJumpIndex;
	    test_reqd = 0;  /* Always return to Action */
	    continue;
	 case RA_ASSIGN:
	 case RA_ASSIGNACT:
	    user_vbls[rp->RuleSelector-FTV1]
	       = rp->RuleMatchedValue[0];  /* Used to be RuleJumpIndex */
	    ++rx;
	    if (rp->RuleAction == RA_ASSIGNACT) test_reqd = 0;
	    continue;
	 case RA_GOSUB:
	 case RA_GOSUBACT:
	    env_stack[e_s_tos++] = rx;  /* Save index of gosub rule */
#ifdef STACK_CHECK
	    if (e_s_tos >= ESTKSIZ) {
	       scpos(0,24);
	       printf("gosub oflo: depth=%d, stk=", e_s_tos);
	       for (j = 0; j != e_s_tos; ++j) printf("%u,", env_stack[j]);
	       printf("\n");
	       return 0;
	       }
#endif
	    rx = rp->RuleJumpIndex;
	    if (rp->RuleAction == RA_GOSUBACT) test_reqd = 0;
	    continue;
	 case RA_GOTO:
	 case RA_GOTOACT:
	    rx = rp->RuleJumpIndex;
#ifdef TESTING
*	    scpos(0,24);  printf("   GOTO %d\n", rx);
#endif
	    if (rp->RuleAction == RA_GOTOACT) test_reqd = 0;
	    continue;
	 case RA_PUSHTO:
	 case RA_PUSHTOACT:
	 case RA_PUSHPKTTO:
	 case RA_PUSHPKTTOACT:
	    a_stack[p_s_depth] = attrib;  /* Remember which rules matched */
	    p_stack[p_s_depth++] = rx;
	    rx = rp->RuleJumpIndex;
	    if (rp->RuleAction == RA_PUSHTOACT ||
		 rp->RuleAction == RA_PUSHPKTTOACT)
	      test_reqd = 0;
#ifdef STACK_CHECK
	    if (p_s_depth >= RSTKSIZ) {
	       scpos(0,24);
	       printf("match oflo: depth=%d, stk=", p_s_depth);
	       for (j = 0; j != p_s_depth; ++j)
		  printf("%u+%u,", a_stack[j],p_stack[j]);
	       printf("\n   LowPeer=");  daddr(from->PeerAddress);
	       printf(", HighPeer=");  daddr(to->PeerAddress);
	       printf("\n");
	       return 0;
	       }
#endif
	    continue;
#ifdef POP_ACT
	 case RA_POPTO:
	 case RA_POPTOACT:
	    --p_s_depth;  /* Forget last matched rule */
	    rx = rp->RuleJumpIndex;
	    if (rp->RuleAction == RA_POPTOACT) test_reqd = 0;
#ifdef STACK_CHECK
	    if (p_s_depth < 0) {
	       scpos(0,24);
	       printf("match uflo: stk=");
	       for (j = 0; j != RSTKSIZ; ++j)
		  printf("%u+%u,", a_stack[j],p_stack[j]);
	       printf("\n   LowPeer=");  daddr(tally->fk.Low.PeerAddress);
	       printf(", HighPeer=");  daddr(tally->fk.High.PeerAddress);
	       printf("\n");
	       return 0;
	       }
#endif
	    continue;
#endif
	    }
	 if (flow_order) {
	    fp->UpOctets += bp->p_len;  fp->UpPDUs += 1;
	    }
	 else {
	    fp->DownOctets += bp->p_len;  fp->DownPDUs += 1;
	    }
	 fp->LastTime = s_uptime;
	 return 1;
	 }
      else ++rx;
      }
   return 0;  /* Not matched */
   }

void pkt_monitor(struct pkt far *bp)
{
   unsigned char j, dtype, dx;
   union decnet far *dp;
   struct rule far *r;
   int len;
#ifdef DECNET
   unsigned int dlen, snode, dnode, area, host;
#endif

   if (bp->PeerAddrType == AT_DUMMY) {
      if (++dummypackets == 1000) {
	 ++kilodummypackets;  dummypackets = 0;
	 }
      ++dummypacketrate;
      }

   /* pkt_match() first param gives value of 'Forward' attribute:
         0 = First test failed, i.e. returned 0
         1 = Performing (or repeating) first test */

   switch (pkt_match(1,0,1, bp, &bp->Low, &bp->High)) {
   case 0:  /* Failed */
      switch (pkt_match(0,0,0, bp, &bp->High, &bp->Low)) {
      case 0:  /* Failed */
      case 1:  /* Matched */
         return;
      case 2:  /* Matched but not yet in count table */
         if (pkt_match(0,0,1, bp, &bp->Low, &bp->High) == 1)
            return;  /* Matched */
         pkt_match(0,1,0, bp, &bp->High, &bp->Low);
         return;
         }
   case 1:  /* Matched */
      return;
   case 2:  /* Matched but not yet in count table */
      if (pkt_match(1,0,0, bp, &bp->High, &bp->Low) == 1)
	 return;  /* Matched */
      pkt_match(1,1,1, bp, &bp->Low, &bp->High);
      return;
      }
   }


#define MAXPKTLEN  1526

#ifdef DECNET
int dn_node(dn_addr)
unsigned char far *dn_addr;
{
   return dn_addr[1]<<8 | dn_addr[0];
   }

#define dn_area(node) (node >> 10)
#define dn_host(node) (node & 0x3FF)
#endif

void unpack_dn_node(ap, dn_addr)
unsigned char far *ap;
unsigned char far *dn_addr;
{
   unsigned char dl = dn_addr[1];
   ap[0] = dl >> 2;    /* DECnet area */
   ap[1] = dl & 0x03;  ap[2] = dn_addr[0];  /* DECnet host */
   ap[3] = 0;
   }

void unpack_at_node(ap, at_net,at_node)
unsigned char far *ap;
unsigned char far *at_net;
unsigned char at_node;
{
   ap[0] = at_net[0];  ap[1] = at_net[1];
   ap[2] = at_node;
   ap[3] = 0;
   }

void pkt_extract(struct pkt huge *bp,
   unsigned char type, unsigned char far *hdrp)
{
   unsigned char j, dtype, dx;
   union decnet far *dp;
#ifdef DECNET
   unsigned int dlen, snode, dnode, area, host;
#endif

   switch (bp->PeerAddrType = type) {
   case AT_IP:
      addrcpy(bp->Low.PeerAddress,&hdrp[12],IP_ADDR_LEN);  /* IP source */
      addrcpy(bp->High.PeerAddress,&hdrp[16],IP_ADDR_LEN);  /* IP dest */
      j = bp->DetailAddrType = hdrp[9];  /* IP Protocol Type */
      if (hdrp[6] & 0x1F != 0 || hdrp[7] != 0) {  /* Fragment offset != 0 */
	 bp->Low.DetailAddress[0] = bp->High.DetailAddress[0] =
	    bp->Low.DetailAddress[1] = bp->High.DetailAddress[1] = 0;
	 return;
	 }
      dx = (hdrp[0] & 0x0F) << 2;  /* HLEN */
	 /* Data offset.  HLEN = IP header length in 32-bit units */
      if (j == PT_ICMP) {
	 bp->Low.DetailAddress[0] = bp->High.DetailAddress[0] = 0;
	 bp->Low.DetailAddress[1] = hdrp[dx];  /* Type field */
	 bp->High.DetailAddress[1] = hdrp[dx+1];  /* Code field */
	 }
      else {
	 bp->Low.DetailAddress[0] = hdrp[dx];  /* Source port */
	 bp->Low.DetailAddress[1] = hdrp[dx+1];
	 bp->High.DetailAddress[0] = hdrp[dx+2];  /* Dest port */
	 bp->High.DetailAddress[1] = hdrp[dx+3];
	 }
      return;
   case AT_NOVELL:
      bp->DetailAddrType = hdrp[5];  /* XNS packet type */
      addrcpy(bp->High.PeerAddress,&hdrp[6],IPX_ADDR_LEN);  /* IPX dest net */
      bp->High.DetailAddress[0] = hdrp[16];
      bp->High.DetailAddress[1] = hdrp[17];  /* IPX dest socket */
      addrcpy(bp->Low.PeerAddress,&hdrp[18],IPX_ADDR_LEN);  /* IPX soure net */
      bp->Low.DetailAddress[0] = hdrp[28];
      bp->Low.DetailAddress[1] = hdrp[29];  /* IPX source socket */
      return;
   case AT_ETHERTALK:
      unpack_at_node(bp->Low.PeerAddress, &hdrp[6],hdrp[9]);  /* AT source */
      unpack_at_node(bp->High.PeerAddress, &hdrp[4],hdrp[8]);  /* AT source */
      bp->DetailAddrType =hdrp[12];  /* AT DDP protocol type */
      bp->Low.DetailAddress[0] = bp->High.DetailAddress[0] = 0;
      bp->Low.DetailAddress[1] = hdrp[11];  /* AT source socket */
      bp->High.DetailAddress[1] = hdrp[10];  /* AT dest socket */
      return;
   case AT_CLNS:
      bp->DetailAddrType = hdrp[5] & 0x1F;  /* CLNS packet type */
      bp->Low.DetailAddress[0] = bp->High.DetailAddress[0] =
	 bp->Low.DetailAddress[1] = bp->High.DetailAddress[1] = 0;
      addrcpy(bp->High.PeerAddress,&hdrp[11],j = hdrp[10]);  /* Dest */
      if (j < NSAP_ADDR_LEN)
	 addrcpy(bp->High.PeerAddress + j, null_flow->fk.High.PeerAddress,
	    NSAP_ADDR_LEN-j);
      addrcpy(bp->Low.PeerAddress,&hdrp[12+j],dx = hdrp[11+j]);  /* Source */
      if (dx < NSAP_ADDR_LEN)
	 addrcpy(bp->Low.PeerAddress + dx, null_flow->fk.Low.PeerAddress,
	    NSAP_ADDR_LEN-dx);
      return;
   case AT_DECNET:
      addrcpy(bp->High.PeerAddress,null_flow->fk.High.PeerAddress,
	 PEER_ADDR_LEN);
      dtype = hdrp[2];  /* DECnet packet type */
      if (dtype != 0x81) dp = (union decnet far *)&hdrp[3];
      else {
	 dtype = hdrp[3];
	 dp = (union decnet far *)&hdrp[4];
	 }
      bp->Low.DetailAddress[0] = bp->High.DetailAddress[0] =
	 bp->Low.DetailAddress[1] = bp->High.DetailAddress[1] = 0;
      switch (bp->DetailAddrType = dtype & 0x0F) {
      case 0x06:  /* Data */
      case 0x0E:  /* Data + discard flag */
	 unpack_dn_node(bp->Low.PeerAddress, dp->d.src_dn_addr);
	 unpack_dn_node(bp->High.PeerAddress, dp->d.dest_dn_addr);
#ifdef DECNET
	 dnode = dn_node(dp->d.dest_dn_addr);
	 snode = dn_node(dp->d.src_dn_addr);
	 fprintf(log,"Data to %d.%d from %d.%d\n",
	    dn_area(dnode),dn_host(dnode),
	    dn_area(snode),dn_host(snode) );
#endif
	 break;
      case 0x07:  /* Level 1 routing */
	 unpack_dn_node(bp->Low.PeerAddress, dp->l1r.src_dn_addr);
#ifdef DECNET
	 snode = dn_node(dp->l1r.src_dn_addr);
	 fprintf(log,"Level 1 routing from %d.%d\n",
	    dn_area(snode),dn_host(snode));
#endif
	 break;
      case 0x09:  /* Level 2 routing */
	 unpack_dn_node(bp->Low.PeerAddress, dp->l2r.src_dn_addr);
#ifdef DECNET
	 snode = dn_node(dp->l2r.src_dn_addr);
	 fprintf(log,"Level 2 routing from %d.%d\n",
	    dn_area(snode),dn_host(snode));
#endif
	 break;
      case 0x0B:  /* Router hello */
	 unpack_dn_node(bp->Low.PeerAddress, dp->rh.src_dn_addr);
#ifdef DECNET
	 snode = dn_node(dp->rh.src_dn_addr);
	 dnode = dn_node(dp->rh.rtr_dn_addr);
	 fprintf(log,"Router hello from %d.%d, other router %d.%d\n",
	    dn_area(snode),dn_host(snode),
	    dn_area(dnode),dn_host(dnode) );
#endif
	 break;
      case 0x0D:  /* Endnode hello */
	 unpack_dn_node(bp->Low.PeerAddress, dp->eh.src_dn_addr);
#ifdef DECNET
	 snode = dn_node(dp->eh.src_dn_addr);
	 dnode = dn_node(dp->eh.rtr_dn_addr);
	 fprintf(log,"Endnode hello from %d.%d, designated router %d.%d\n",
	    dn_area(snode),dn_host(snode),
	    dn_area(dnode),dn_host(dnode) );
#endif
	 break;
      default:  /* Unknown DECnet type */
	 scpos(0,24);
	 printf("\nDN pkt type %02x: ", dtype);
	 for (j=0; j != 16; ++j) printf(" %02x",hdrp[j]);
	 printf("\n");
	 for (j=16; j != 34; ++j) printf(" %02x",hdrp[j]);
	 printf("\n");
#ifdef TESTING
	 if (!log) open_log();
	 fprintf(log, "\nDN pkt type %02x: ", dtype);
	 for (j=17; j != 34; ++j) fprintf(log," %02x",hdrp[j]);
	 fprintf(log, "\n");
#endif
	 }
      return;
      }
   }

void other_extract(struct pkt huge *bp,
   unsigned int ethertype, unsigned int lsap)
{
   bp->PeerAddrType = AT_OTHER;
   putshort(bp->Low.PeerAddress,ethertype);  /* Blue Book type */
   putshort(bp->High.PeerAddress,lsap);  /* 802.2 LSAP */
   bp->DetailAddrType =
      bp->Low.DetailAddress[0] = bp->Low.DetailAddress[1] =
      bp->High.DetailAddress[0] = bp->High.DetailAddress[1] = 0;
   }

void handle_pkt(struct pkt huge *pp,
   unsigned int ether_type, unsigned int lsap,
   unsigned int pktlen, unsigned char far *ethp, unsigned char far *p)
{
   pp->PeerAddrType = AT_DUMMY;
   switch (ether_type) {
   case 0x0800:
      if (proto_reqd[AT_IP]) pkt_extract(pp, AT_IP, p);
      break;
   case 0x809B:
      if (proto_reqd[AT_ETHERTALK]) pkt_extract(pp, AT_ETHERTALK, p);
      break;
   case 0x8137:
      if (proto_reqd[AT_NOVELL]) pkt_extract(pp, AT_NOVELL, p);
      break;
   case 0x6003:
      if (proto_reqd[AT_DECNET]) pkt_extract(pp, AT_DECNET, p);
      break;
   case 0x0000:  /* 802.3 packet other than SNAP */
      switch (lsap) {
      case 0xFFFF:
         if (proto_reqd[AT_NOVELL]) pkt_extract(pp, AT_NOVELL, p);
         break;  /* Novell "Raw 802.2" doesn't use LLC */
#ifdef CLNS
      case 0xFEFE:  /* CLNS */
         if (proto_reqd[AT_CLNS]) pkt_extract(pp, AT_CLNS, p+2);
         break;  /* ISO 8473-1 uses 1-org, byte 0 is LLC control byte */
#endif
      case 0xE0E0:  /* Novell 802.2 */
         if (proto_reqd[AT_NOVELL]) pkt_extract(pp, AT_NOVELL, p+3);
         break;
/*    case 0xF0F0:  NetBIOS */
         }
      }
   if (proto_reqd[AT_OTHER] && pp->PeerAddrType == AT_DUMMY)
      other_extract(pp, ether_type,lsap);

   if (pp->PeerAddrType != AT_DUMMY && adj_reqd) {
      addrcpy(pp->Low.AdjAddress,ethp+6,MAC_ADDR_LEN);  /* Source */
      addrcpy(pp->High.AdjAddress,ethp,MAC_ADDR_LEN);  /* Dest */
      }

   pp->p_len = pktlen;
   }

#ifdef AU_MSDOS
void save_time()
{
   s_tod_h = tod_h;  s_tod_m = tod_m;  s_tod_s = tod_s;
   elapsed_sec = 0;
   }
#endif

#ifdef DEBUG
void write_flows()
{
   int x;
   char msg[40];
   struct flow far *t;
   if (!log) open_log();
   for (x = 1; x <= nflows; ++x) {
      if ((t = find_flow(x)) == NULL) continue;
      sprintf(msg,"flow %d:  ",x);  pkey(msg,t);
      fprintf(log,"   Up: %8lu %6lu Down: %8lu %6lu Time: %8lu %8lu\n",
	 t->UpOctets,t->UpPDUs, t->DownOctets,t->DownPDUs,
	 t->FirstTime, t->LastTime);
      }
   fclose(log);  log = NULL;
   }
#endif

unsigned int active_flows()
{
   unsigned int af,j;
   struct flow far *fp;
   for (af = 0, j = 2; j <= mxflowsp1; ++j) {
      /* Flow indexes from 2 to MXFLOWS+1 */
      if ((fp = flow_ix[j-1]) == NULL)  /* Unused flow index */
	 continue;
      if (fp->LastTime > GarbageCollectTime)  /* Can't be recovered */
	 ++af;
      }
   return af;
   }

void show_flow_table()
{
   char msg[60];
   unsigned int j;
   unsigned long f;
   struct CTinfo *ctip;
   display_msg(1,"Flow Table Info ..");
   if (n_collectors == 0)
      display_msg(0,"No collections so far");
   else {
      display_msg(0,"Collection times:");
      for (j = 0; j != n_collectors; ++j) {
	 ctip = &CTi[j];
	 sprintf(msg, "  %u.%u.%u.%u  %5lu s, %5lu s",
	    ctip->LastCollectPeer[0],ctip->LastCollectPeer[1],
	    ctip->LastCollectPeer[2],ctip->LastCollectPeer[3],
	    (s_uptime-ctip->PrevCollectTime)/100L,
	    (s_uptime-ctip->LastCollectTime)/100L);
	 }
      display_msg(0,msg);
      }
   sprintf(msg, "InactTime %lu, HighWater %u, Flood %u",
      InactivityTimeout,HighWaterMark,FloodMark);
   display_msg(0,msg);
   sprintf(msg, "CurrentRuleSet %u, EmergencyRuleSet %u",
      CurrentRuleSet,EmergencyRuleSet);
   display_msg(0,msg);
   sprintf(msg, "Flows: %u active, %u used, %u max",
      active_flows(), nflows, mxflowsp1-1);
   display_msg(0,msg);
   sprintf(msg,"  %lu recovered (GC: %u s, %u flow)",
      FlowsRecovered, gc_interval,gc_f);
   display_msg(0,msg);
   if (stats_time != 0) {
      f = FlowsCreated*1000L/stats_time;
      sprintf(msg,"  Creating   %lu.%03lu flow/s", f/1000,f%1000);
      display_msg(0,msg);
      f = FlowsRecovered*1000L/stats_time;
      sprintf(msg,"  Recovering %lu.%03lu flow/s", f/1000,f%1000);
      display_msg(0,msg);
      }
   }

void show_protocols()
{
   int j;
   char msg[40];
   display_msg(0,"Protocols being metered ..");
   sprintf(msg,"      %3.3s  %s",
      adj_reqd ? "Yes" : " No", "Adjacent");
   display_msg(0,msg);
   for (j = 0; j != sizeof(proto_reqd); ++j) {
      if (proto_names[j][0] != '\0') {
	 sprintf(msg,"  %2d  %3.3s  %s",
	    j, proto_reqd[j] ? "Yes" : " No", proto_names[j]);
	 display_msg(0,msg);
	 }
      }
   }

void zero_stats(void)
{
   FlowsCreated = FlowsRecovered =
      n_matches = n_hash_compares = n_hash_searches = 0L;
   clear_pkt_stats = 1;
   display_msg(1,"Statistics Zeroed");
   }

void show_stats(void)
{
   unsigned int i,j;
   unsigned long aps,apb, kdp;
   char msg[60];
#ifdef AU_MSDOS
   if (stats_time == 0 || npackets == 0 ||
	 kilodummypackets == 0 || mindummyrate == 0)
#else
   if (stats_time == 0 || npackets == 0)
#endif
      return;  /* Avoid divides by zero */

   display_msg(1,"Meter Statistics ..");
   aps = (npackets*10+5L)/(stats_time*10L);
#ifdef AU_MSDOS
   apb = (t_backlog*10+5L)/(stats_time*10L);
   sprintf(msg,"Av pkt/s %lu, av pkt backlog %lu", aps,apb);
   display_msg(0,msg);
   sprintf(msg,"Max pkt/s %u, max pkt backlog %u",
      max_pkt_rate,max_pkt_backlog);
   display_msg(0,msg);
   if (kilodummypackets != 0) {
      if (dummypackets >= 500) ++kilodummypackets;
      i = kilodummypackets*1000L/(kilodummypackets+npackets/1000L);
      j = (mindummyrate*10000L+5L)/((mindummyrate+mdpacketrate)*10L);
      sprintf(msg,"Idle time av %u.%u, min %u.%u %", i/10,i%10, j/10,j%10);
      display_msg(0,msg);
      }
#else
   sprintf(msg,"Av pkt/s %lu, max pkt/s %u", aps,max_pkt_rate);
   display_msg(0,msg);
#endif
   sprintf(msg,"%u flows active (max %u)",active_flows(),mxflowsp1-1);
   display_msg(0,msg);
   i = (n_matches*100L+5L)/(npackets*10L);
   j = (n_hash_searches*100L+5L)/(npackets*10L);
   sprintf(msg,"%u.%u rules/pkt, %u.%u counts/pkt",
      i/10,i%10, j/10,j%10);
   display_msg(0,msg);
   if (n_hash_searches != 0) {
      i = (n_hash_compares*100L+5L)/(n_hash_searches*10L);
      sprintf(msg,"%u.%u compares/count", i/10,i%10);
      display_msg(0,msg);
      sprintf(msg,"%u hash slots, %u in use, ", t_hash_size,n_hash_ents);
      }
   }

void init_monitor()
{
   int j;
   unsigned char far *fp;
   unsigned char *np;
#ifdef AU_MSDOS
   set_tod();
   save_time();  /* For screen display */
   start_uptime_clock();  /* Starts 1 tick early; flows have CreateTime > 1 */
#else
   boot_time = 0;
#endif
   fp = (unsigned char far *)(null_flow = get_flow());
   for (j = 0; j != sizeof(struct flow); ++j) *fp++ = 0;
   tally = get_flow();
   rule_hash = (int far *)farmalloc(sizeof(int)*(rh_size = INITHSZ));
   flow_ix = (struct flow far **)malloc(sizeof(struct flow far *)*(mxflowsp1));
   for (j = 0; j != mxflowsp1; ++j) flow_ix[j] = NULL;
   nflows = 0;  /* No accounting flows in use yet */
   gcf_ix = empty_ix = mxflowsp1;  /* Flow 2 will be the first allocated */
   np = (unsigned char *)CTi;
   for (j = 0; j != sizeof(CTi); ++j) *np++ = 0;
   n_collectors = 0;
   s_uptime = uptime();

   /* Start with default rule tables */

   rule_table[0] = (struct rule far *)default_rule_table;
   rt_size[0] = DRT_SIZE;  ct_size[0] = DCT_SIZE;
   for (j = 1; j != MXRTBLS; ++j) {
      rule_table[j] = NULL;
      rt_size[j] = ct_size[j] = 0;
      }
   for (j = 0; j != MXCTBLS; ++j) count_table[j] = NULL;
   open_rule_set(1, 1); /* Open set 1 */
   }

void show_help()
{
   display_msg(0,"Copyright (C) 1992-1995 by Nevil Brownlee");
   display_msg(0,"Computer Centre, The University of Auckland");
   display_msg(0,"");
   display_msg(0,"Keyboard commands ..");
   display_msg(0,"");
#ifdef AU_MSDOS
   display_msg(0,"  b: show Bad packet counts");
   display_msg(0,"  m: show Memory usage");
#endif
   display_msg(0,"  p: show Protocols");
   display_msg(0,"  s: show Statistics");
   display_msg(0,"  v: show meter Version");
   display_msg(0,"  z: Zero statistics");
   if (kb_enabled) {
      display_msg(0,"");
      display_msg(0,"Esc: stop metering, exit NeTraMet");
      }
   }

void handle_kb(ch)
int ch;
{
   switch (tolower(ch)) {
   case 'f':
      show_flow_table();
      break;
   case 'p':
      show_protocols();
      break;
   case 's':
      show_stats();
      break;
   case 't':
      show_meter_time();
      break;
#ifdef DEBUG
   case 'x':
      if (kb_enabled) write_flows();
      break;
#endif
   case 'z':
      zero_stats();
      break;
   case '?':
      show_help();
      break;
   default:
      break;
      }
   }
