/* 1640, Wed 7 Jun 95

   FLOWHASH.H:  Data structures for AU Meter

   Copyright (C) 1992-1995 by Nevil Brownlee,
   Computer Centre,  The University of Auckland */

#ifndef EXTFLOW
#define EXTFLOW  extern
#define DECLARE  0
#define INIT(v)
#else
#define EXTFLOW
#define DECLARE  1
#define INIT(v)  = v
#endif

#ifdef BIG_PEER_ADDR
#define DFMXFLOWS 2600  /* Default max nbr of traffic flows */
#else
#define DFMXFLOWS 3200  /* Default max nbr of traffic flows */
#endif

#define MXCLCTRS    10  /* Max nbr of active Collectors */
#define MXRTBLS     10  /* Max nbr of rule tables */
#define MXCTBLS     50  /* Max nbr of count tables */

#define RSTKSIZ     50  /* Max depth of rule matches stack */
#define ESTKSIZ     50  /* Max depth of rule gosub stack */
#define MXRHTSZ   1024  /* Max size for rule-group hash table */
#define HASHMOD    512  /* Size of flow hash tables */
#define MNCGRPSZ     5  /* Min nbr of rules in a compare group */
#define INITHSZ   1500  /* Initial rule_hash size (unsigned shorts) */
#define FLOWBLOBSZ  55  /* Nbr of {index, info} objects in a column blob */

struct hash_tbl {
   struct flow far *hash_ent[HASHMOD];
   };
   
#define AT_IGNORE     0  /* Addr_type values */
#define AT_DUMMY      1
#define AT_ADJACENT   1

#define AT_IP         2
#define AT_CLNS       3
#define AT_IDPR       4
#define AT_DECNET     5
#define AT_NOVELL     6
#define AT_ETHERTALK  7
#define AT_DETAIL     8
#define AT_OTHER     12

#define PT_ICMP       1  /* IP protocol types */
#define PT_TCP        6
#define PT_UDP       17
#define PT_OSPF      89

EXTFLOW char addr_len[]  /* Lengths for each address type */
#if DECLARE
   = {
      PEER_ADDR_LEN,    /*  0 Default */
      MAC_ADDR_LEN,     /*  1 Adjacent */
      IP_ADDR_LEN,      /*  2 IP */
      NSAP_ADDR_LEN,    /*  3 CLNS */
      4,                /*  4 IDPR */
      DN_ADDR_LEN,      /*  5 DECnet */
      IPX_ADDR_LEN,     /*  6 Novell IPX */
      ET_ADDR_LEN,      /*  7 EtherTalk */
      DETAIL_ADDR_LEN,  /*  8 Detail */
      0,	        /*  9 */
      0,	        /* 10 */
      0,                /* 11 */
      DETAIL_ADDR_LEN   /* 12 Other */
      }
#endif
   ;

EXTFLOW char *proto_names[]
#if DECLARE
   = {
      "",              /*  0 Default */
      "",              /*  1 Adjacent */
      "IP",            /*  2 IP */
#ifdef CLNS
      "CLNS",          /*  3 CLNS */
#else
      "",              /*  3 CLNS */
#endif
      "",              /*  4 IDPR */
      "DECnet IV",     /*  5 DECnet */
      "Novell IPX",    /*  6 Novell IPX */
      "EtherTalk",     /*  7 EtherTalk */
      "",              /*  8 Detail */
      "",              /*  9 */
      "",              /* 10 */
      "",              /* 11 */
      "Other"          /* 12 Other */
      }
#endif
   ;

EXTFLOW char proto_reqd[1+AT_OTHER];  /* 1 => protocol tested for by rules */
EXTFLOW char adj_reqd;  /* 1 => adjacent addresses tested for by rules */

struct key {
   unsigned char AdjAddress[MAC_ADDR_LEN], AdjMask[MAC_ADDR_LEN];
   unsigned char PeerAddress[PEER_ADDR_LEN], PeerMask[PEER_ADDR_LEN];
   unsigned char DetailAddress[DETAIL_ADDR_LEN], DetailMask[DETAIL_ADDR_LEN];
   unsigned char Class, Kind;
   };

struct flow_key {
   unsigned char PeerAddrType, DetailAddrType;
   struct key Low, High;
   unsigned char FlowClass, FlowKind;
   };

struct flow {
   union {  /* Must be first to line up with hash_tbl entries */
      struct flow far *action;     /* For aggregates and tallies */
      struct rule far *count;      /* For counts */
      struct hash_tbl huge *hash;  /* For tally action entries */
      struct flow far *next;       /* For hash chains */
      } link;
   unsigned long UpOctets,UpPDUs, DownOctets,DownPDUs,
      FirstTime,LastTime;
   unsigned char FlowRuleSet;
   struct flow_key fk;
   };

EXTFLOW unsigned int nflows,  /* Nbr of flows in use */
   mxflowsp1;  /* Max nbr of flows + 1 */
EXTFLOW struct flow far **flow_ix;  /* The flow table */
   /* Flow 1 is a dummy for the Create and Active tables */
   
EXTFLOW unsigned int empty_ix;  /* Last flow number used by number_flow */

EXTFLOW unsigned int gcf_ix;  /* Next possible flow for garbage collector */

EXTFLOW unsigned char  /* Parameters for garbage collector */
   gc_interval INIT(10),  /* Seconds between calls */
   gc_f INIT(4);  /* Flow indexes to test each call */
EXTFLOW unsigned long GarbageCollectTime INIT(0),
   /* OK to recover flows with LastTime greater than this */
   FlowsCreated INIT(0),
   FlowsRecovered INIT(0);

struct rule {
   unsigned char  RuleSelector;  /* Rule's MIB variables */
   unsigned char  RuleMask[RULE_ADDR_LEN];
   unsigned char  RuleMatchedValue[RULE_ADDR_LEN];
   unsigned char  RuleAction;
   unsigned int   RuleJumpIndex;

   unsigned short hash_tbl_index;  /* 0 => simple compare */
   unsigned short hash_link;
   };

EXTFLOW struct flow far *null_flow;  /* All fields zeroed */

EXTFLOW struct hash_tbl huge *count_table[MXCTBLS];
EXTFLOW unsigned int ct_size[MXRTBLS];
EXTFLOW unsigned int c_csz;  /* Nbr of counts in current rule table */

EXTFLOW struct rule far *rule_table[MXRTBLS];
EXTFLOW unsigned int rt_size[MXRTBLS];
EXTFLOW struct rule huge *c_rt;  /* Current Rule Table */
EXTFLOW unsigned int c_rsz;  /* Nbr of rules in current rule table */

struct rule_hash_tbl {
   unsigned short group_last;  /* Nbr of last rule in group */
   unsigned short hash_mask;  /* Group hash table size - 1 */
   unsigned short hash_ent[1];
   };

EXTFLOW int huge *rule_hash;
EXTFLOW unsigned int rh_size;

#define RA_IGNORE        1  /* RuleAction values */
#define RA_RETRY         2
#define RA_COUNT         3
#define RA_COUNTPKT      4
#define RA_RETURN        5
#define RA_GOSUB         6
#define RA_GOSUBACT      7
#define RA_ASSIGN        8
#define RA_ASSIGNACT     9
#define RA_GOTO         10
#define RA_GOTOACT      11
#define RA_PUSHTO       12
#define RA_PUSHTOACT    13
#define RA_PUSHPKTTO    14
#define RA_PUSHPKTTOACT 15
#ifdef POP_ACT
#define RA_POPTO        16
#define RA_POPTOACT     17
#endif

#define RS_NULL                 0  /* Rule Selector dummy value */

#define FTFLOWINDEX		1  /* Flow table attribute values */
#define FTFLOWSTATUS		2

#define FTLOWINTERFACE          3
#define FTLOWADJACENTTYPE       4
#define FTLOWADJACENTADDRESS    5
#define FTLOWADJACENTMASK       6
#define FTLOWPEERTYPE		7
#define FTLOWPEERADDRESS	8
#define FTLOWPEERMASK	        9
#define FTLOWDETAILTYPE        10
#define FTLOWDETAILADDRESS     11
#define FTLOWDETAILMASK        12

#define FTHIINTERFACE          13
#define FTHIADJACENTTYPE       14
#define FTHIADJACENTADDRESS    15
#define FTHIADJACENTMASK       16
#define FTHIPEERTYPE	       17
#define FTHIPEERADDRESS	       18
#define FTHIPEERMASK	       19
#define FTHIDETAILTYPE         20
#define FTHIDETAILADDRESS      21
#define FTHIDETAILMASK         22

#define FTPDUSCALE             23
#define FTOCTETSCALE           24
#define FTRULESET              25

#define FTUPOCTETS	       26
#define FTUPPDUS	       27
#define FTDOWNOCTETS	       28
#define FTDOWNPDUS	       29
#define FTFIRSTTIME	       30
#define FTLASTTIME	       31

#define FTLOWSUBSCRIBERID      32
#define FTHISUBSCRIBERID       33
#define FTSESSIONID            34

#define FTSOURCECLASS	       35
#define FTDESTCLASS            36
#define FTFLOWCLASS	       37
#define FTSOURCEKIND 	       38
#define FTDESTKIND 	       39
#define FTFLOWKIND 	       40

#define NATTRIBS     FTFLOWKIND

#define FTFORWARD    50
#define FTV1         51
#define FTV2         52
#define FTV3         53
#define FTV4         54
#define FTV5         55

#define NBRVBLS  (FTV5-FTFORWARD)  /* Nbr of user variables */

struct attrib_info {
   char *name;
   unsigned char index;
   unsigned char len;
   };

EXTFLOW struct attrib_info attribs[1+NATTRIBS]
   /* Serial search -> longest names first! */
#if DECLARE
   = {
      "null",                   0,                     0,
      "flowindex",              FTFLOWINDEX,           2,
      "flowstatus",             FTFLOWSTATUS,          1,
      "sourceinterface",        FTLOWINTERFACE,        1,
      "sourceadjacenttype",     FTLOWADJACENTTYPE,     1,
      "sourceadjacentaddress",  FTLOWADJACENTADDRESS,  MAC_ADDR_LEN,
      "sourceadjacentmask",     FTLOWADJACENTMASK,     MAC_ADDR_LEN,
      "sourcepeertype",         FTLOWPEERTYPE,         1,
      "sourcepeeraddress",      FTLOWPEERADDRESS,      PEER_ADDR_LEN,
      "sourcepeermask",         FTLOWPEERMASK,         PEER_ADDR_LEN,
      "sourcetranstype",        FTLOWDETAILTYPE,       1,
      "sourcetransaddress",     FTLOWDETAILADDRESS,    DETAIL_ADDR_LEN,
      "sourcetransmask",        FTLOWDETAILMASK,       DETAIL_ADDR_LEN,

      "destinterface",          FTHIINTERFACE,         1,
      "destadjacenttype",       FTHIADJACENTTYPE,      1,
      "destadjacentaddress",    FTHIADJACENTADDRESS,   MAC_ADDR_LEN,
      "destadjacentmask",       FTHIADJACENTMASK,      MAC_ADDR_LEN,
      "destpeertype",           FTHIPEERTYPE,          1,
      "destpeeraddress",        FTHIPEERADDRESS,       PEER_ADDR_LEN,
      "destpeermask",           FTHIPEERMASK,          PEER_ADDR_LEN,
      "desttranstype",          FTHIDETAILTYPE,        1,
      "desttransaddress",       FTHIDETAILADDRESS,     DETAIL_ADDR_LEN,
      "desttransmask",          FTHIDETAILMASK,        DETAIL_ADDR_LEN,

      "pduscale",               FTPDUSCALE,            0,
      "octetscale",             FTOCTETSCALE,          0,
      "flowruleset",            FTRULESET,             1,
      "tooctets",               FTUPOCTETS,            4,
      "topdus",                 FTUPPDUS,              4,
      "fromoctets",             FTDOWNOCTETS,          4,
      "frompdus",               FTDOWNPDUS,            4,
      "firsttime",              FTFIRSTTIME,           4,
      "lasttime",               FTLASTTIME,            4,

      "sourcesubscriberid",     FTLOWSUBSCRIBERID,     0,
      "destsubscriberid",       FTHISUBSCRIBERID,      0,
      "sessionid",              FTSESSIONID,           0,

      "sourceclass",		FTSOURCECLASS,	       1,
      "destclass",		FTDESTCLASS,           1,
      "flowclass",		FTFLOWCLASS,	       1,
      "sourcekind",		FTSOURCEKIND,	       1,
      "destkind",		FTDESTKIND,            1,
      "flowkind",		FTFLOWKIND,	       1
   }
#endif
   ;

/* Global Variables */

/* Objects in a column blob start with two bytes of flow number.
   Fixed-size objects, e.g. DetailAddresses (2 bytes), AddressTypes
      (1 byte), follow the flow nbr, i.e. (flownbr(2), object(n)).
   Variable-size objects, e.g. PeerAddresses and PeerMasks, have one
      size byte preceding them, and have their trailing zeros truncated,
      i.e. (flownbr(2), size(1), object(size)) */

#define FB_OBJECT_SZ  (2+1+MAC_ADDR_LEN)
#define FB_DATA_SZ    FLOWBLOBSZ*FB_OBJECT_SZ

EXTFLOW unsigned char column_blob[FB_DATA_SZ + 2*(2+1+RULE_ADDR_LEN)];

#define GET_SHORT(a)  (a[0] << 8 | a[1])
unsigned char *putshort(unsigned char far *d, unsigned int n);
unsigned char *putlong(unsigned char far *d, unsigned long n);

EXTFLOW unsigned int n_hash_tables INIT(0),  /* Performance statistics */
   n_hash_ents INIT(0), t_hash_size INIT(0);
EXTFLOW unsigned long n_matches INIT(0),
   n_hash_compares INIT(0), n_hash_searches INIT(0);

EXTFLOW unsigned long s_uptime;  /* uptime() value for current sample */

EXTFLOW unsigned char kb_enabled;
EXTFLOW unsigned char display_enabled;

/* MIB variables */

EXTFLOW unsigned long InactivityTimeout INIT(600L);  /* 10 minutes */
EXTFLOW unsigned char HighWaterMark INIT(65);  /* 65 % */
EXTFLOW unsigned char FloodMark INIT(95);  /* 95 % */

struct CTinfo {
   unsigned char LastCollectPeer[PEER_ADDR_LEN];
   unsigned long LastCollectTime;
   unsigned long PrevCollectTime;
   };

EXTFLOW struct CTinfo CTi[MXCLCTRS];
EXTFLOW unsigned char n_collectors INIT(0);

EXTFLOW unsigned long LastCollectTime INIT(0L);

EXTFLOW unsigned char CurrentRuleSet INIT(0);
EXTFLOW unsigned char EmergencyRuleSet INIT(0);

EXTFLOW unsigned long pcap_LostPackets INIT(0);
EXTFLOW unsigned long noflowpackets INIT(0);

EXTFLOW unsigned char AdjAddrType;

/* Meter outer block functions */

void one_second_process(void);

/* Flowhash functions */

void pkt_extract(struct pkt huge *bp,
   unsigned char type, unsigned char far *hdrp);
void other_extract(struct pkt huge *bp,
   unsigned int ethertype, unsigned int lsap);
void handle_pkt(struct pkt huge *pp,
   unsigned int ether_type, unsigned int lsap,
   unsigned int pktlen, unsigned char far *ethp, unsigned char far *p);

#define addrcpy(d,s,n) qmove(s,d,n)  /* Stick to d,s order for copies! */

struct flow far *get_flow(void);
void free_flow(struct flow far *q);

unsigned int rule_table_counts(unsigned char t);
int open_rule_set(unsigned char doSet, unsigned char n);
void close_rule_set(void);

unsigned int number_flow(struct flow far *fp);
int garbage_collect(unsigned char incremental);
void more_garbage(void);
unsigned int active_flows(void);

struct flow far *find_flow(int x);

struct flow far *search_hash_table(
   struct hash_tbl far *t,  /* Table to search */
   struct pkt_key far *Low,  /* Target keys */
   struct pkt_key far *High,
   struct pkt far *bp,
   unsigned char OK_to_add  /* 1 = add new flow if not found */
   );

int pkt_match(unsigned char forward, unsigned char OK_to_match,
   unsigned char flow_order,
   struct pkt far *bp,
   struct pkt_key far *from, struct pkt_key far *to);

void build_flow_blob(unsigned long t_time, unsigned int fx);

void pkt_monitor(struct pkt far *bp);

void open_log(void);
void write_flows(void);

void show_stats(void);
void show_time(void);

void init_monitor(void);

void show_help(void);
void handle_kb(int ch);

